import logging
from cobrakbase.core.kbasegenomesgenome import normalize_role

logger = logging.getLogger(__name__)

class TemplateManipulator:
    def __init__(self, template):
        self.template = template
    
    def get_role_to_rxn_ids(self):
        role_to_rxn_ids = {}
        for o in self.template.data['reactions']:
            trxn = self.template.get_reaction(o['id'])
            for role_id in trxn.get_roles():
                if not role_id in role_to_rxn_ids:
                    role_to_rxn_ids[role_id] = set()
                role_to_rxn_ids[role_id].add(trxn.id)
        
        return role_to_rxn_ids
    
    def get_search_name_to_role_id(self):
        # get duplicates
        search_name_to_role_id = {}
        for role in self.template.data['roles']:
            n = normalize_role(role['name'])
            if not n in search_name_to_role_id:
                search_name_to_role_id[n] = set()
            search_name_to_role_id[n].add(role['id'])
        return search_name_to_role_id
    
    def clear_orphan_roles(self):
        role_to_rxn_ids = self.get_role_to_rxn_ids()
        used_roles = set(role_to_rxn_ids)
        search_name_to_role_id = self.get_search_name_to_role_id()
    
        to_delete = set()
        to_update = set()
        for sn in search_name_to_role_id:
            role_ids = list(sorted(search_name_to_role_id[sn]))
            if len(search_name_to_role_id[sn]) > 1:
                selected = []
                for role_id in role_ids:
                    rxn_count = len(role_to_rxn_ids[role_id]) if role_id in role_to_rxn_ids else 0
                    if rxn_count > 0:
                        selected.append(role_id)
                if len(selected) == 0:
                    selected.append(role_ids[0])
                if len(selected) == 1:
                    s = selected[0]
                    role = self.template.get_role(s)
                    to_update.add(s)
                    for role_id in role_ids:
                        if not role_id == s:
                            to_delete.add(role_id)
                            other_role = self.template.get_role(role_id)
                            role['aliases'].append('alias:' + other_role['name'])
                else:
                    logger.warning('unable to select role %s', role_ids)

        self.template.data['roles']= list(filter(lambda x : x['id'] not in to_delete, self.template.data['roles']))
        
        return to_update, to_delete
    
    def clean_template(self, source_allow):
        search_name_to_role_id = self.get_search_name_to_role_id()

        # clear roles
        role_source = {}
        roles_to_clear = set()
        seed_roles = set()
        for o in self.template.data['roles']:
            role_source[o['id']] = o['source']
            if not o['source'] == source_allow:
                roles_to_clear.add(o['id'])
            else:
                seed_roles.add(o['id'])

        complexes_to_clear = set()
        for complex_o in self.template.data['complexes']:
            #complex_o
            complex_source_from_role = set()
            for complex_role in complex_o['complexroles']:
                role_id = complex_role['templaterole_ref'].split('/')[-1]
                complex_source_from_role.add(role_source[role_id])
            if len(complex_source_from_role) == 0:
                #print(complex_o)
                pass
            elif len(complex_source_from_role) == 1:
                if not list(complex_source_from_role)[0] == complex_o['source']:
                    ss = list(complex_source_from_role)[0]
                    logger.debug('%s %s', ss, complex_o['source'])
                    complex_o['source'] = ss
            else:
                print(complex_o)
            if not complex_o['source'] == source_allow:
                complexes_to_clear.add(complex_o['id'])

        reactions_to_clear = set()
        for reaction_o in self.template.data['reactions']:
            templatecomplex_refs = reaction_o['templatecomplex_refs']
            templatecomplex_refs_filter = list(filter(
                lambda x : not x.split('/')[-1] in complexes_to_clear, 
                templatecomplex_refs))
            if not len(templatecomplex_refs_filter) == len(templatecomplex_refs):
                logger.debug('[%s] removing complexes', reaction_o['id'])
                reaction_o['templatecomplex_refs'] = templatecomplex_refs_filter
            if len(templatecomplex_refs) == 0:
                reactions_to_clear.add(reaction_o['id'])

        template_reactions_filter = list(filter(lambda x : not x['id'] in reactions_to_clear, self.template.data['reactions']))

        return template_reactions_filter

def add_role_rxn(trxn, functions):
    for function_id in functions:
        n = annotation_api.neo4j_graph.nodes[int(function_id)]
        nfunction = Neo4jAnnotationFunction(n)
        sn = normalize_role(nfunction.value)
        role_id = function_id
        if sn in search_name_to_role_id:
            role_id = search_name_to_role_id[sn]
        if len(nfunction.sub_functions) == 0:
            #check if role in rxn
            cpx_roles = trxn.get_complex_roles()
            match = dict(filter(lambda x : role_id in x[1], cpx_roles.items()))
            if len(match) == 0:
                template_cpx_id = template.get_complex_from_role(set([role_id]))
                if template_cpx_id == None:
                    new_role_id, template_cpx_id = template.add_solo_role_with_complex(nfunction.value)
                    #print('+', trxn.id, new_role_id, template_cpx_id, role_id, nfunction.value)
                #trxn.data['templatecomplex_refs'].append('~/complexes/id/' + template_cpx_id)

def report(annotation_api, user_id):
    for rxn_id in a['fungi']:
        for function_id in a['fungi'][rxn_id]['current']:
            action = a['fungi'][rxn_id]['current'][function_id]
            if user_id in a['fungi'][rxn_id]['user'][int(function_id)]:
                role_str = annotation_api.matcher.get(int(function_id))['key']
                ss = "{}\t{}\t{}\t{}".format(rxn_id, role_str, action, a['fungi'][rxn_id]['user'][int(function_id)])
                print(ss)

def get_curation_data_(doc):
    
    functions = doc['functions']
    function_user_data = {}
    for l in doc['log']:
        target = l['target']
        user_id = l['user_id']
        t = l['timestamp']
        action = l['action']
        if not target in function_user_data:
            function_user_data[target] = {}
        if not user_id in function_user_data[target]:
            function_user_data[target][user_id] = (action, t)
        else:
            prev_t = function_user_data[target][user_id][1]
            if prev_t < t:
                function_user_data[target][user_id] = (action, t)
    return functions, function_user_data

def get_curation_data(template_set_id, a):
    accept_scores = set([
        #'opt_score3', #*
        #'opt_score2', #**
        'opt_score1', #***
    ])
    remove = {}
    accept = {}
    for rxn_id in a[template_set_id]:
        remove[rxn_id] = set()
        accept[rxn_id] = set()
        mapping = a[template_set_id][rxn_id]['current']
        for function_id in mapping:
            if mapping[function_id] in accept_scores:
                accept[rxn_id].add(function_id)
            else:
                remove[rxn_id].add(function_id)
                
    return accept, remove

def get_function(function_id, annotation_api):
    n = annotation_api.neo4j_graph.nodes[int(function_id)]
    nfunction = Neo4jAnnotationFunction(n)
    sn = normalize_role(nfunction.value)
    role_id = function_id
    if sn in search_name_to_role_id:
        role_id = search_name_to_role_id[sn]
    #print(function_id, role_id, nfunction, sn)
    return nfunction, role_id

def remove_role(trxn, role_id):
    complex_roles = trxn.get_complex_roles()
    delete_set = set()
    for complex_id in complex_roles:
        if role_id in complex_roles[complex_id]:
            delete_set.add('~/complexes/id/' + complex_id)

    templatecomplex_refs = list(filter(
        lambda x : not x in delete_set, 
        trxn.data['templatecomplex_refs']))
    
    return templatecomplex_refs

def add_role(trxn, role_ids):
    complex_roles = trxn.get_complex_roles()
    role_match = {}
    for o in role_ids:
        role_match[o] = False
    for complex_id in complex_roles:
        for o in role_match:
            if o in complex_roles[complex_id]:
                role_match[o] = True
    all_roles_present = True
    for o in role_match:
        all_roles_present &= role_match[o]
    if all_roles_present:
        logger.debug('ignore %s all present in atleast 1 complex', role_ids)
        return None
    complex_id = template.get_complex_from_role(role_ids)
    if complex_id == None:
        logger.warning('unable to find complex for %s', role_ids)
        return None
    complex_ref = '~/complexes/id/' + complex_id
    if complex_ref in trxn.data['templatecomplex_refs']:
        logger.debug('already contains complex %s, role %s', role_ids, complex_ref)
        return None
    return complex_ref

def get_role_change(rxn_id, test_accept, test_remove):
    rxn_role_change = {}
    for function_id in test_accept[rxn_id]:
        rxn_role_change[function_id] = True
    for function_id in test_remove[rxn_id]:
        if not function_id in rxn_role_change:
            rxn_role_change[function_id] = False
        else:
            print('!')
    return rxn_role_change

def update_roles(trxn, rxn_role_change):
    #trxn_roles = trxn.get_complex_roles()
    #for cpd_id in trxn_roles:
    #    for role_id in trxn_roles[cpd_id]:
    #        logger.warning('%s %s %s', rxn_id, role_id, role_id_to_name[role_id])
    for function_id in rxn_role_change:
        nfunction, role_id = get_function(function_id)
        if len(nfunction.sub_functions) == 0:
            if len(role_id) > 1 and rxn_role_change[function_id]:
                logger.warning('%s multiple role ids %s %s', rxn_id, nfunction, role_id)
            else:
                role_id = list(role_id)[0]
                role = template.get_role(role_id)
                if rxn_role_change[function_id]:
                    #print('accept or add')
                    templatecomplex_ref = add_role(trxn, [role_id])
                    if not templatecomplex_ref == None:
                        trxn.data['templatecomplex_refs'].append(templatecomplex_ref)
                    logger.debug('%s accept or add role %s, %s', rxn_id, nfunction, templatecomplex_ref)
                else:
                    #print('delete or ignore', role_id, role['name'])
                    templatecomplex_refs = remove_role(trxn, role_id)
                    if not templatecomplex_refs == None:
                        trxn.data['templatecomplex_refs'] = templatecomplex_refs
                    logger.debug('%s delete role %s, removed %d complex(es)', rxn_id, nfunction, len(trxn.data['templatecomplex_refs']) - len(templatecomplex_refs))                
                #print(('+' if rxn_role_change[function_id] else '-') + function_id, nfunction, role_id, role['source'])
        else:
            logger.debug('%s ignore multifunction role', rxn_id)

#'/Users/fliu/workspace/jupyter/web/annotation/data/TempModels/MitoCore.json'
def aaaaaa(f):
    mmm = None
    with open(f, 'r') as fh:
        mmm = json.loads(fh.read())
    for o in mmm['metabolites']:
        o['id'] += '@MitoCore'
    for o in mmm['reactions']:
        o['id'] += '@MitoCore'
        o['metabolites'] = dict(map(lambda x : (x[0] + '@MitoCore', x[1]), o['metabolites'].items()))

    with open(f, 'w') as fh:
        fh.write(json.dumps(mmm))


import re
from cobrakbase.core.kbaseobject import KBaseObjectBase
from cobrakbase.core.utils import get_id_from_ref


class KBaseGenomeFeature(KBaseObjectBase):
    
    @property
    def dna_sequence(self):
        return self.data['dna_sequence']
    
    @property
    def protein_translation(self):
        return self.data['protein_translation']
    
    def __str__(self):
        return self.data['id']
    
    @property
    def functions(self):
        return KBaseGenomeFeature.split_annotation(self.functions_unsplit)
    
    @property
    def functions_unsplit(self):
        functions = set([None])
        if 'functions' in self.data:
            if type(self.data['functions']) == str:
                return set([self.data['functions']])
            functions = set(self.data['functions'])
        elif 'function' in self.data:
            functions = set([self.data['function']])
        return functions
    
    @staticmethod
    def split_annotation(annotation):
        split = set()
        for s in annotation:
            if not s == None:
                for comp in re.split(' / | @ |; ', s):
                    split.add(comp)
            else:
                split.add(s)
        return split

    @staticmethod
    def split_function(function):
        split = set()
        if not function == None:
            for comp in re.split(' / | @ |; ', function):
                split.add(comp)
        else:
            split.add(s)
        return split


import logging

logger = logging.getLogger(__name__)

class KBaseFBABuilder():
    
    def __init__(self, flux_dist):
        self.flux_dist = flux_dist
        self.id = 'fba1'
        self.model = None
        self.media = None
        self.objective_value = None
        self.ws = None
    
    @staticmethod
    def fromCobra(object_id, model, solution, media, ws):
        flux_dist = dict(solution.fluxes)
        b = KBaseFBABuilder(flux_dist)
        b.id = object_id
        b.model = model
        b.media = media
        b.ws = ws
        b.objective_value = solution.objective_value
        return b
    
    def withFvaSolution(self, fva):
        pass
    
    @property
    def fbamodel_ref(self):
        #65434/6/1 filipeliu:narrative_1592958114519/model1
        return "{}/{}".format(self.ws, self.model.id)
    
    @property
    def media_ref(self):
        #65434/5/1 filipeliu:narrative_1592958114519/Carbon-D-Glucose
        return "{}/{}".format(self.ws, self.media.data['name'])
    
    def build(self):
        #Saving final solution as an FBA object in KBase
        fbaobj = {
            "FBABiomassVariables": [
                {
                    "biomass_ref": "~/fbamodel/biomasses/id/bio1",
                    "class": "unknown",
                    "lowerBound": self.model.reactions.bio1_biomass.lower_bound,
                    "upperBound": self.model.reactions.bio1_biomass.upper_bound,
                    "max": self.model.reactions.bio1_biomass.upper_bound,
                    "min": self.model.reactions.bio1_biomass.lower_bound,
                    "other_max": [],
                    "other_min": [],
                    "other_values": [],
                    "value": self.flux_dist['bio1_biomass'],
                    "variableType": "biomassflux"
                }
            ],"FBACompoundBounds": [],
            "FBACompoundVariables": [],
            "FBAConstraints": [],
            "FBADeletionResults": [],
            "FBAMetaboliteProductionResults": [],
            "FBAMinimalMediaResults": [],
            "FBAMinimalReactionsResults": [],
            "FBAPromResults": [],
            "FBAReactionBounds": [],
            "FBAReactionVariables": [],
            "FBATintleResults": [],
            "MFALog": "",
            "PROMKappa": 1,
            "QuantitativeOptimizationSolutions": [],
            "__VERSION__": 1,
            "additionalCpd_refs": [],
            "allReversible": 0,
            "biomassRemovals": {},
            "biomassflux_objterms": {
                "bio1": 1
            },
            "calculateReactionKnockoutSensitivity": 0,
            "comboDeletions": 0,
            "compoundflux_objterms": {},
            "decomposeReversibleDrainFlux": 0,
            "decomposeReversibleFlux": 0,
            "defaultMaxDrainFlux": 0,
            "defaultMaxFlux": 1000,
            "defaultMinDrainFlux": -1000,
            "drainfluxUseVariables": 0,
            "fbamodel_ref": self.fbamodel_ref,
            "findMinimalMedia": 0,
            "fluxMinimization": 1,
            "fluxUseVariables": 0,
            "fva": 0,
            "gapfillingSolutions": [],
            "geneKO_refs": [],
            "id": self.id,
            "inputfiles": {},
            "maximizeActiveReactions": 0,
            "maximizeObjective": 1,
            "media_list_refs": [],
            "media_ref": self.media_ref,
            "minimizeErrorThermodynamicConstraints": 0,
            "minimize_reaction_costs": {},
            "minimize_reactions": 0,
            "noErrorThermodynamicConstraints": 0,
            "numberOfSolutions": 1,
            "objectiveConstraintFraction": 0.1,
            "objectiveValue": self.objective_value,
            "other_objectives": [],
            "outputfiles": {},
            "parameters": {
                "Auxotrophy metabolite list": "",
                "Beachhead metabolite list": "",
                "minimum_target_flux": "0.01",
                "save phenotype fluxes": "0",
                "suboptimal solutions": "1"
            },
            "quantitativeOptimization": 0,
            "reactionKO_refs": [],
            "reactionflux_objterms": {},
            "simpleThermoConstraints": 0,
            "thermodynamicConstraints": 0,
            "uptakeLimits": {}
        }
        
        for varname, value in self.flux_dist.items():
            rxn = self.model.reactions.get_by_id(varname)

            if varname.startswith("EX_"):
                fbaobj["FBACompoundVariables"].append({
                    "class": "Variable",
                    "lowerBound": rxn.lower_bound,
                    "max": rxn.upper_bound,
                    "min": rxn.lower_bound,
                    "upperBound": rxn.upper_bound,
                    "modelcompound_ref": "~/fbamodel/modelcompounds/id/"+varname[3:],
                    "other_max": [],
                    "other_min": [],
                    "other_values": [],
                    "value": -1*value,
                    "variableType": "drainflux"
                })
            elif varname.endswith("_biomass") or varname.startswith("DM_"):
                logger.debug("SKIP %s", varname)
            else:
                fbaobj["FBAReactionVariables"].append({
                    "biomass_dependencies": [],
                    "class": "unknown", #needs FVA
                    "coupled_reactions": [],
                    "exp_state": "unknown",
                    "expression": 0,
                    "lowerBound": rxn.lower_bound,
                    "max": rxn.upper_bound,
                    "min": rxn.lower_bound,
                    "upperBound": rxn.upper_bound,
                    "modelreaction_ref": "~/fbamodel/modelreactions/id/"+varname,
                    "other_max": [],
                    "other_min": [],
                    "other_values": [],
                    "scaled_exp": 0,
                    "value": value,
                    "variableType": "flux"
                })
        return fbaobj


from cobrakbase.core.kbaseobject import KBaseObjectBase
from cobrakbase.core.utils import get_id_from_ref

class NewModelTemplate(KBaseObjectBase):
    
    

    def new(id, name):
        return {
            'biochemistry_ref': "489/6/6",
            'biomasses': [],
            'compartments': [],
            'compcompounds': [],
            'complexes': [],
            'compounds': [],
            'domain': [],
            'id': id,
            'name': name,
            'pathways': [],
            'reactions': [],
            'roles': [],
            'subsystems': [],
            'type': "GenomeScale",
        }


import logging
from cobrakbase.core.kbasegenomesgenome import normalize_role
logger = logging.getLogger(__name__)

class TemplateCuration:
    
    def __init__(self, template, curation_api, annotation_api):
        self.template = template
        self.curation_api = curation_api
        self.annotation_api = annotation_api
        
    def get_function_user_data(self, doc):
        functions = doc['functions']
        function_user_data = {}
        for l in doc['log']:
            target = l['target']
            user_id = l['user_id']
            t = l['timestamp']
            action = l['action']
            if not target in function_user_data:
                function_user_data[target] = {}
            if not user_id in function_user_data[target]:
                function_user_data[target][user_id] = (action, t)
            else:
                prev_t = function_user_data[target][user_id][1]
                if prev_t < t:
                    function_user_data[target][user_id] = (action, t)
        return functions, function_user_data
    
    def get_reaction_annotation(self):
        a = {}
        for doc in self.curation_api['templates_reactions'].find():
            rxn_id, template_id = doc['_id'].split('@')
            functions, function_user_data = self.get_function_user_data(doc)
            if not template_id in a:
                a[template_id] = {}
            a[template_id][rxn_id] = {
                'current' : functions,
                'user' : function_user_data
            }
        return a
    
    def get_curation_data(self, template_set_id):
        a = self.get_reaction_annotation()
        accept_scores = set([
            #'opt_score3', #*
            #'opt_score2', #**
            'opt_score1', #***
        ])
        remove = {}
        accept = {}
        for rxn_id in a[template_set_id]:
            remove[rxn_id] = set()
            accept[rxn_id] = set()
            mapping = a[template_set_id][rxn_id]['current']
            for function_id in mapping:
                if mapping[function_id] in accept_scores:
                    accept[rxn_id].add(function_id)
                else:
                    remove[rxn_id].add(function_id)

        return accept, remove
    
    def get_function(self, function_id, search_name_to_role_id):
        nfunction = self.annotation_api.get_function_by_uid(int(function_id))
        sn = normalize_role(nfunction.value)
        #sn = nfunction.search_value
        role_id = function_id
        if sn in search_name_to_role_id:
            role_id = search_name_to_role_id[sn]
        #print(function_id, role_id, nfunction, sn)
        return nfunction, role_id

    def get_roles_to_add(self, test_accept, search_name_to_role_id):
        accept_role_uids = set()
        for rxn_id in test_accept:
            for role_uid in test_accept[rxn_id]:
                accept_role_uids.add(role_uid)
                
        roles_to_add = set()
        for role_uid in accept_role_uids:
            nfunction, role_id = self.get_function(role_uid, search_name_to_role_id)
            if not type(role_id) == set:
                roles_to_add.add(nfunction.value)
            elif len(role_id) == 1:
                roles_to_add.add(nfunction.value)
            else:
                print(role_id, list(map(lambda x : self.template.get_role(x)['source'], role_id)))
                
        sn_to_roles = {}
        for role_name in roles_to_add:
            sn = normalize_role(role_name)
            if not sn in sn_to_roles:
                sn_to_roles[sn] = set()
            sn_to_roles[sn].add(role_name)
                
        return sn_to_roles
    
    def get_role_change(self, rxn_id, test_accept, test_remove):
        rxn_role_change = {}
        if rxn_id in test_accept:
            for function_id in test_accept[rxn_id]:
                rxn_role_change[function_id] = True
        if rxn_id in test_remove:
            for function_id in test_remove[rxn_id]:
                if not function_id in rxn_role_change:
                    rxn_role_change[function_id] = False
                else:
                    print('!')
        return rxn_role_change
    
    def add_role(self, trxn, role_ids, auto_complex = False):
        complex_roles = trxn.get_complex_roles()
        role_match = {}
        for o in role_ids:
            role_match[o] = False
        for complex_id in complex_roles:
            for o in role_match:
                if o in complex_roles[complex_id]:
                    role_match[o] = True
        all_roles_present = True
        for o in role_match:
            all_roles_present &= role_match[o]
        if all_roles_present:
            logger.debug('ignore %s all present in atleast 1 complex', role_ids)
            return None
        complex_id = self.template.get_complex_from_role(role_ids)
        if complex_id == None:
            logger.warning('unable to find complex for %s', role_ids)
            if auto_complex:
                role_names = set()
                for role_id in role_ids:
                    role = self.template.get_role(role_id)
                    role_names.add(role['name'])
                logger.warning('build complex for %s', role_names)
                complex_id = self.template.add_complex_from_role_names(role_names)
            else:
                return None
        complex_ref = '~/complexes/id/' + complex_id
        if complex_ref in trxn.data['templatecomplex_refs']:
            logger.debug('already contains complex %s, role %s', role_ids, complex_ref)
            return None
        return complex_ref
    
    def remove_role(self, trxn, role_id):
        complex_roles = trxn.get_complex_roles()
        delete_set = set()
        for complex_id in complex_roles:
            if role_id in complex_roles[complex_id]:
                delete_set.add('~/complexes/id/' + complex_id)

        templatecomplex_refs = list(filter(
            lambda x : not x in delete_set, 
            trxn.data['templatecomplex_refs']))

        return templatecomplex_refs
    
    def update_roles(self, trxn, rxn_role_change, search_name_to_role_id, auto_complex = False):
        #trxn_roles = trxn.get_complex_roles()
        #for cpd_id in trxn_roles:
        #    for role_id in trxn_roles[cpd_id]:
        #        logger.warning('%s %s %s', rxn_id, role_id, role_id_to_name[role_id])
        for function_id in rxn_role_change:
            nfunction, role_id = self.get_function(function_id, search_name_to_role_id)
            if type(role_id) == int or type(role_id) == str:
                logger.warning('%s function not in template [%s] %s', trxn.id, nfunction, role_id)
            elif len(nfunction.sub_functions) == 0:
                if len(role_id) > 1 and rxn_role_change[function_id]:
                    logger.warning('%s multiple role ids %s %s', trxn.id, nfunction, role_id)
                else:
                    role_id = list(role_id)[0]
                    role = self.template.get_role(role_id)
                    if rxn_role_change[function_id]:
                        #print('accept or add')
                        templatecomplex_ref = self.add_role(trxn, [role_id], auto_complex)
                        if not templatecomplex_ref == None:
                            trxn.data['templatecomplex_refs'].append(templatecomplex_ref)
                        logger.debug('%s accept or add role %s, %s', trxn.id, nfunction, templatecomplex_ref)
                    else:
                        #print('delete or ignore', role_id, role['name'])
                        templatecomplex_refs = self.remove_role(trxn, role_id)
                        if not templatecomplex_refs == None:
                            trxn.data['templatecomplex_refs'] = templatecomplex_refs
                        logger.debug('%s delete role %s, removed %d complex(es)', trxn.id, nfunction, len(trxn.data['templatecomplex_refs']) - len(templatecomplex_refs))                
                    #print(('+' if rxn_role_change[function_id] else '-') + function_id, nfunction, role_id, role['source'])
            else:
                logger.debug('%s ignore multifunction role', trxn.id)


import cobrakbase.core.kbasefba
from cobrakbase.core.model import KBaseFBAModel
from cobrakbase.core.kbasebiochemmedia import KBaseBiochemMedia
from cobrakbase.core.kbasefbafba import KBaseFBA
from cobrakbase.core.kbasegenomesgenome import KBaseGenome


import json

def read_json(filename):
    data = None
    with open(filename, 'r') as f:
        data = json.loads(f.read())
    return data

def write_json(data, filename, pretty=False):
    with open(filename, 'w') as f:
        if pretty:
            f.write(json.dumps(data, indent=4, sort_keys=True))
        else:
            f.write(json.dumps(data))


from cobrakbase.core.kbasefba.newmodeltemplate import NewModelTemplate
from cobrakbase.core.kbasefba.template_curation import TemplateCuration
from cobrakbase.core.kbasefba.template_manipulator import TemplateManipulator


from cobrakbase.core.kbaseobject import KBaseObjectBase
from cobrakbase.core.utils import get_id_from_ref
from cobrakbase.core.kbasegenomesgenome import normalize_role

class TemplateRole(KBaseObjectBase):
    
    def __init__(self, data, template = None):
        super().__init__(data)
        self.template = template
    

class NewTemplateReaction(KBaseObjectBase):
    
    def __init__(self, data, template = None):
        super().__init__(data)
        self.template = template
    
    def remove_role():
        pass
    
    def get_roles(self):
        res = set()
        for complexes in self.data['templatecomplex_refs']:
            complex_id = complexes.split('/')[-1]
            cpx = self.template.get_complex(complex_id)

            for complexrole in cpx['complexroles']:
                role_id = complexrole['templaterole_ref'].split('/')[-1]
                res.add(role_id)
                
        return res
    
    def get_complexes(self):
        res = set()
        for complexes in self.data['templatecomplex_refs']:
            complex_id = complexes.split('/')[-1]
            res.add(complex_id)
            
        return res
    
    def get_complex_roles(self):
        res = {}
        for complexes in self.data['templatecomplex_refs']:
            complex_id = complexes.split('/')[-1]
            res[complex_id] = set()
            cpx = self.template.get_complex(complex_id)

            for complexrole in cpx['complexroles']:
                role_id = complexrole['templaterole_ref'].split('/')[-1]
                res[complex_id].add(role_id)
        return res

class NewModelTemplate(KBaseObjectBase):
    
    def __init__(self, data, role_suf = 'ftr', complex_suf = 'cpx'):
        super().__init__(data)
        self.role_suf = role_suf
        self.complex_suf = complex_suf
        self.role_last_id = self.get_last_id_value(self.data['roles'], role_suf)
        self.complex_last_id = self.get_last_id_value(self.data['complexes'], complex_suf)
        
        self.role_set_to_cpx = {}
        self.search_name_to_role_id = {}
        for role in self.data['roles']:
            self.search_name_to_role_id[normalize_role(role['name'])] = role['id']
        for cpx in self.data['complexes']:
            roles = set()
            for complexrole in cpx['complexroles']:
                role_id = complexrole['templaterole_ref'].split('/')[-1]
                roles.add(role_id)
            #print(cpx, roles)
            self.role_set_to_cpx[';'.join(sorted(roles))] = cpx['id']
        
    def get_role_sources(self):
        pass
    
    def get_complex_sources(self):
        pass
        
    def get_complex_from_role(self, roles):
        cpx_role_str = ';'.join(sorted(roles))
        if cpx_role_str in self.role_set_to_cpx:
            return self.role_set_to_cpx[cpx_role_str]
        return None
        
    def add_role(self, name, source = 'ModelSEED'):
        sn = normalize_role(name)
        if sn in self.search_name_to_role_id:
            return self.search_name_to_role_id[sn]
        self.role_last_id += 1
        role_id = self.role_suf + str(self.role_last_id).zfill(5)
        self.data['roles'].append({
            'aliases': [],
            'features': [],
            'id': role_id,
            'name': name,
            'source': source
        })
        self.search_name_to_role_id[sn] = role_id
        return role_id
    
    def add_complex_from_role_names(self, role_names, source = 'ModelSEED'):
        role_ids = set(map(lambda o : self.add_role(o), role_names))
        complex_id = self.get_complex_from_role(role_ids)
        if complex_id == None:
            self.complex_last_id += 1
            complex_id = self.complex_suf + str(self.complex_last_id).zfill(5)
            complex_data = {
                'complexroles': [],
                'confidence': 0,
                'id': complex_id,
                'name': complex_id,
                'reference': 'null',
                'source': source
            }
            for role_id in role_ids:
                complex_data['complexroles'].append({
                    'optional_role': 0,
                    'templaterole_ref': '~/roles/id/' + role_id,
                    'triggering': 1
                })
            
            self.data['complexes'].append(complex_data)
            self.role_set_to_cpx[';'.join(sorted(role_ids))] = complex_id
            
        return complex_id
        
    def add_solo_role_with_complex(self, name):
        role_id = self.add_role(name)
        complex_id = self.add_complex_from_role_names([name])
        return role_id, complex_id
        
    def get_last_id_value(self, l, s):
        last_id = 0
        for o in l:
            id = o['id']
            if id.startswith(s):
                number_part = id[len(s):]
                if len(number_part) == 5:
                    if int(number_part) > last_id:
                        last_id = int(number_part)
        return last_id
    
    def get_complex(self, id):
        res = list(filter(lambda x : x['id'] == id, self.data['complexes']))
        if len(res) == 0:
            return None
        return res[0]
    
    def get_reaction(self, id):
        res = list(filter(lambda x : x['id'] == id, self.data['reactions']))
        if len(res) == 0:
            return None
        return NewTemplateReaction(res[0], self)
    
    def get_role(self, id):
        res = list(filter(lambda x : x['id'] == id, self.data['roles']))
        if len(res) == 0:
            return None
        return res[0]
    
    